package be.iminds.ilabt.jfed.highlevel.util;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.util.common.RFC3339Util;
import org.jetbrains.annotations.Contract;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.text.ParseException;
import java.time.Instant;
import java.util.Collection;
import java.util.Date;
import java.util.List;

public class ExpirationChecker {
    @Nullable
    public static Instant getEarliestExpiration(@Nonnull Collection<Sliver> slivers) {
        Instant earliestSliverExpiration = null;
        for (Sliver sliver: slivers) {
            if (sliver.getExpirationDate() != null && !sliver.isFake()) {
                if (earliestSliverExpiration == null ||
                        earliestSliverExpiration.isAfter(sliver.getExpirationDate()))
                    earliestSliverExpiration = sliver.getExpirationDate();
            }
        }
        return earliestSliverExpiration;
    }

    @Nullable
    public static Instant getEarliestExpiration(@Nonnull Slice slice) {
        Instant earliestSliverExp = getEarliestSliverExpiration(slice);
        if (earliestSliverExp == null)
            return null;
        if (slice.getExpirationDate() == null)
            return null;
        if (slice.getExpirationDate().isBefore(earliestSliverExp)) {
            return slice.getExpirationDate();
        } else {
            return earliestSliverExp;
        }
    }

    @Nullable
    public static Instant getEarliestSliverExpiration(@Nonnull Slice slice) {
        List<Sliver> slivers = slice.getSliversCopy();
        Instant earliestSliverExp = getEarliestExpiration(slivers);
        // not sure if using this is a good idea: slice.getStatus().getEarliestKnownExpires()
        if (earliestSliverExp == null) {
            if (slivers.isEmpty() || slivers.stream().allMatch(Sliver::isFake)) {
                //very nasty: in AMv3, only RSpec has info on expire date when there are no (non-fake) slivers
                if (slice.getManifestRspec() != null) {
                    ModelRspec modelRspec = slice.getManifestRspec().getImmutableModelRspec();
                    if (modelRspec != null && modelRspec.getExpiresString() != null) {
                        try {
                            return RFC3339Util.rfc3339StringToInstant(modelRspec.getExpiresString());
                        } catch (ParseException e) {
                            return null;
                        }
                    }
                }
                return null;
//                return slice.getExpirationDate();  //we don't know, so better return null than just the slice date
            } else {
                return null;
            }
        }
        return earliestSliverExp;
    }

    @Nullable
    public static Instant getEarliestExpiration(@Nonnull Experiment experiment) {
        return getEarliestExpiration(experiment.getSlice());
    }

    @Nullable
    @Contract("!null,!null -> !null;_,null->null;null,_->null")
    public static Instant getEarliestOrNull(@Nullable Instant a, @Nullable Instant b) {
        if (a == null || b == null) {
            return null;
        }
        return a.isBefore(b) ? a : b;
    }

    @Nullable
    @Contract("!null,!null -> !null;null,null->null;null,!null->!null;!null,null->!null")
    public static Instant getEarliestNonNull(@Nullable Instant a, @Nullable Instant b) {
        if (a == null) { return b; }
        if (b == null) { return a; }
        return a.isBefore(b) ? a : b;
    }

    @Contract("_,null -> true")
    public static boolean isEarlyExpire(@Nonnull Instant requestedExpiration, @Nullable Instant actualExpiration) {
        if (actualExpiration == null)
            return true;
        if (requestedExpiration == null)
            throw new IllegalStateException("requestedExpiration is null");

        //Request times are rounded down to the nearest second. We have to take that into account.
        //We also don't care about, say, 5 seconds, so we'll just ignore up to a 5 second difference.
        long actualMillis = actualExpiration.toEpochMilli();
        long requestedMillis = requestedExpiration.toEpochMilli();
        return (actualMillis + 5000L) < requestedMillis;
    }

    @Contract("_,null -> true")
    public static boolean isEarlyExpire(@Nonnull Experiment experiment, @Nullable Instant timeToCheck) {
        return isEarlyExpire(experiment.getRequestedEndTime(), timeToCheck);
    }

    public static boolean isEarlyExpire(@Nonnull Experiment experiment, @Nonnull Slice slice) {
        return isEarlyExpire(experiment.getRequestedEndTime(), getEarliestExpiration(slice));
    }

    public static boolean isEarlyExpire(@Nonnull Experiment experiment) {
        return isEarlyExpire(experiment.getRequestedEndTime(), getEarliestExpiration(experiment.getSlice()));
    }
}
